#include <string.h>
#include <intrin.h>

static unsigned int k[] = {
	0x428a2f98, 0x71374491, 0xb5c0fbcf, 0xe9b5dba5, 0x3956c25b, 0x59f111f1, 0x923f82a4, 0xab1c5ed5,
	0xd807aa98, 0x12835b01, 0x243185be, 0x550c7dc3, 0x72be5d74, 0x80deb1fe, 0x9bdc06a7, 0xc19bf174,
	0xe49b69c1, 0xefbe4786, 0x0fc19dc6, 0x240ca1cc, 0x2de92c6f, 0x4a7484aa, 0x5cb0a9dc, 0x76f988da,
	0x983e5152, 0xa831c66d, 0xb00327c8, 0xbf597fc7, 0xc6e00bf3, 0xd5a79147, 0x06ca6351, 0x14292967,
	0x27b70a85, 0x2e1b2138, 0x4d2c6dfc, 0x53380d13, 0x650a7354, 0x766a0abb, 0x81c2c92e, 0x92722c85,
	0xa2bfe8a1, 0xa81a664b, 0xc24b8b70, 0xc76c51a3, 0xd192e819, 0xd6990624, 0xf40e3585, 0x106aa070,
	0x19a4c116, 0x1e376c08, 0x2748774c, 0x34b0bcb5, 0x391c0cb3, 0x4ed8aa4a, 0x5b9cca4f, 0x682e6ff3,
	0x748f82ee, 0x78a5636f, 0x84c87814, 0x8cc70208, 0x90befffa, 0xa4506ceb, 0xbef9a3f7, 0xc67178f2
};

static unsigned int hash[8], w[64];
static unsigned long total_size;

static void process_buffer()
{
	unsigned int a, b, c, d, e, f, g, h, i;
	unsigned int s0, s1, ch, maj, temp1, temp2;

	for (i = 0; i < 16; i++)
		w[i] = __builtin_bswap32(w[i]);
	for (i = 16; i < 64; i ++) {
		s0 = _rotr(w[i - 15], 7) ^ _rotr(w[i - 15], 18) ^ w[i - 15] >> 3;
		s1 = _rotr(w[i - 2], 17) ^ _rotr(w[i - 2], 19) ^ w[i - 2] >> 10;
		w[i] = w[i - 16] + s0 + w[i - 7] + s1;
	}
	a = hash[0];
	b = hash[1];
	c = hash[2];
	d = hash[3];
	e = hash[4];
	f = hash[5];
	g = hash[6];
	h = hash[7];
	for (i = 0; i < 64; i++) {
		s1 = _rotr(e, 6) ^ _rotr(e, 11) ^ _rotr(e, 25);
		ch = (e & f) ^ (~e & g);
		temp1 = h + s1 + ch + k[i] + w[i];
		s0 = _rotr(a, 2) ^ _rotr(a, 13) ^ _rotr(a, 22);
		maj = (a & b) ^ (a & c) ^ (b & c);
		temp2 = s0 + maj;
		h = g;
		g = f;
		f = e;
		e = d + temp1;
		d = c;
		c = b;
		b = a;
		a = temp1 + temp2;
	}
	hash[0] += a;
	hash[1] += b;
	hash[2] += c;
	hash[3] += d;
	hash[4] += e;
	hash[5] += f;
	hash[6] += g;
	hash[7] += h;
}

void sha256(void *in, unsigned long inlen, void *out, _Bool init, _Bool finalize)
{
	int i;

	if (init) {
		hash[0] = 0x6a09e667;
		hash[1] = 0xbb67ae85;
		hash[2] = 0x3c6ef372;
		hash[3] = 0xa54ff53a;
		hash[4] = 0x510e527f;
		hash[5] = 0x9b05688c;
		hash[6] = 0x1f83d9ab;
		hash[7] = 0x5be0cd19;
		total_size = inlen;
	} else
		total_size += inlen;

	while (inlen >= 64) {
		memcpy(w, in, 64);
		process_buffer();
		inlen -= 64;
		in += 64;
	}

	if (finalize) {
		memcpy(w, in, inlen);
		*((unsigned char*)w + inlen) = 0x80;
		memset((unsigned char*)w + inlen + 1, 0, 63 - inlen);
		if (inlen > 64 - 9) {
			process_buffer();
			memset(w, 0, 56);
		}
		total_size <<= 3;
		for (i = 63; i >= 56; i--) {
			((unsigned char*)w)[i] = total_size;
			total_size >>= 8;
		}
		process_buffer();
		for (i = 0; i < 8; i++)
			((unsigned int*)out)[i] = __builtin_bswap32(hash[i]);
	}
}
