#include <intrin.h>

static unsigned char buffer[64];
static unsigned long long *hash;
static unsigned long value;

static unsigned long long C[8][256];

static unsigned char sbox[256] = {
	0x18, 0x23, 0xc6, 0xe8, 0x87, 0xb8, 0x01, 0x4f, 0x36, 0xa6, 0xd2, 0xf5, 0x79, 0x6f, 0x91, 0x52,
	0x60, 0xbc, 0x9b, 0x8e, 0xa3, 0x0c, 0x7b, 0x35, 0x1d, 0xe0, 0xd7, 0xc2, 0x2e, 0x4b, 0xfe, 0x57,
	0x15, 0x77, 0x37, 0xe5, 0x9f, 0xf0, 0x4a, 0xda, 0x58, 0xc9, 0x29, 0x0a, 0xb1, 0xa0, 0x6b, 0x85,
	0xbd, 0x5d, 0x10, 0xf4, 0xcb, 0x3e, 0x05, 0x67, 0xe4, 0x27, 0x41, 0x8b, 0xa7, 0x7d, 0x95, 0xd8,
	0xfb, 0xee, 0x7c, 0x66, 0xdd, 0x17, 0x47, 0x9e, 0xca, 0x2d, 0xbf, 0x07, 0xad, 0x5a, 0x83, 0x33,
	0x63, 0x02, 0xaa, 0x71, 0xc8, 0x19, 0x49, 0xd9, 0xf2, 0xe3, 0x5b, 0x88, 0x9a, 0x26, 0x32, 0xb0,
	0xe9, 0x0f, 0xd5, 0x80, 0xbe, 0xcd, 0x34, 0x48, 0xff, 0x7a, 0x90, 0x5f, 0x20, 0x68, 0x1a, 0xae,
	0xb4, 0x54, 0x93, 0x22, 0x64, 0xf1, 0x73, 0x12, 0x40, 0x08, 0xc3, 0xec, 0xdb, 0xa1, 0x8d, 0x3d,
	0x97, 0x00, 0xcf, 0x2b, 0x76, 0x82, 0xd6, 0x1b, 0xb5, 0xaf, 0x6a, 0x50, 0x45, 0xf3, 0x30, 0xef,
	0x3f, 0x55, 0xa2, 0xea, 0x65, 0xba, 0x2f, 0xc0, 0xde, 0x1c, 0xfd, 0x4d, 0x92, 0x75, 0x06, 0x8a,
	0xb2, 0xe6, 0x0e, 0x1f, 0x62, 0xd4, 0xa8, 0x96, 0xf9, 0xc5, 0x25, 0x59, 0x84, 0x72, 0x39, 0x4c,
	0x5e, 0x78, 0x38, 0x8c, 0xd1, 0xa5, 0xe2, 0x61, 0xb3, 0x21, 0x9c, 0x1e, 0x43, 0xc7, 0xfc, 0x04,
	0x51, 0x99, 0x6d, 0x0d, 0xfa, 0xdf, 0x7e, 0x24, 0x3b, 0xab, 0xce, 0x11, 0x8f, 0x4e, 0xb7, 0xeb,
	0x3c, 0x81, 0x94, 0xf7, 0xb9, 0x13, 0x2c, 0xd3, 0xe7, 0x6e, 0xc4, 0x03, 0x56, 0x44, 0x7f, 0xa9,
	0x2a, 0xbb, 0xc1, 0x53, 0xdc, 0x0b, 0x9d, 0x6c, 0x31, 0x74, 0xf6, 0x46, 0xac, 0x89, 0x14, 0xe1,
	0x16, 0x3a, 0x69, 0x09, 0x70, 0xb6, 0xd0, 0xed, 0xcc, 0x42, 0x98, 0xa4, 0x28, 0x5c, 0xf8, 0x86
};

static unsigned long long rc[10] = {
	0x4f01b887e8c62318,
	0x52916f79f5d2a636,
	0x357b0ca38e9bbc60,
	0x57fe4b2ec2d7e01d,
	0xda4af09fe5377715,
	0x856ba0b10a29c958,
	0x67053ecbf4105dbd,
	0xd8957da78b4127e4,
	0x9e4717dd667ceefb,
	0x33835aad07bf2dca
};

void whirlpool_ini()
{
	int i, j;
	union {unsigned char ch[16]; unsigned long long ll[2];} v;

	for (i = 0; i < 256; i++) {
		v.ch[6] = (v.ch[0] = v.ch[1] = v.ch[3] = sbox[i]) << 1;
		if (v.ch[0] & 0x80) v.ch[6] ^= 0x1d;
		v.ch[2] = v.ch[6] << 1;
		if (v.ch[6] & 0x80) v.ch[2] ^= 0x1d;
		v.ch[5] = v.ch[2] ^ v.ch[0];
		v.ch[4] = v.ch[2] << 1;
		if (v.ch[2] & 0x80) v.ch[4] ^= 0x1d;
		v.ch[7] = v.ch[4] ^ v.ch[0];
		v.ll[1] = v.ll[0];
		for (j = 0; j < 8; j++)
			C[j][i] = *(unsigned long long*)((unsigned char*)&v + 8 - j);
	}
}

__attribute__((optimize("-O1")))
static void processBuffer()
{
	union {unsigned char ch[128]; unsigned long long ll[16];} K, state;
	union {unsigned char ch[64]; unsigned long long ll[8];} L;
	int r, i, j;

	for (i = 0; i < 8; i++) L.ll[i] = (K.ll[i + 8] = K.ll[i] = hash[i]) ^ ((unsigned long long*)buffer)[i];

	for (r = 0; r < 10; r++) {
		for (i = j = 0; i < 64; i += 8, j++) {
			state.ll[j + 8] = state.ll[j] = L.ll[j];
			L.ll[j] = C[0][K.ch[i]] ^ C[1][K.ch[i + 57]] ^ C[2][K.ch[i + 50]] ^ C[3][K.ch[i + 43]] ^ C[4][K.ch[i + 36]] ^ C[5][K.ch[i + 29]] ^ C[6][K.ch[i + 22]] ^ C[7][K.ch[i + 15]];
		}
		L.ll[0] ^= rc[r];
		for (i = j = 0; i < 64; i += 8, j++) {
			K.ll[j + 8] = K.ll[j] = L.ll[j];
			L.ll[j] ^= C[0][state.ch[i]] ^ C[1][state.ch[i + 57]] ^ C[2][state.ch[i + 50]] ^ C[3][state.ch[i + 43]] ^ C[4][state.ch[i + 36]] ^ C[5][state.ch[i + 29]] ^ C[6][state.ch[i + 22]] ^ C[7][state.ch[i + 15]];
		}
	}

	for (i = 0; i < 8; i++) hash[i] ^= L.ll[i] ^ ((unsigned long long*)buffer)[i];
}

void whirlpool(void *in, unsigned long inlen, void *out, _Bool init, _Bool finalize)
{
	unsigned long bufferPos, sourcePos;

	hash = out;
	if (init) {
		memset(hash, 0, 64);
		value = inlen * 8;
	} else
		value += inlen * 8;

	sourcePos = 0;
	while (sourcePos + 64 <= inlen) {
		memcpy(buffer, in + sourcePos, 64);
		processBuffer();
		sourcePos += 64;
	}

	if (finalize) {
		bufferPos = inlen - sourcePos;
		memcpy(buffer, in + sourcePos, bufferPos);
		buffer[bufferPos++] = 0x80;
		if (bufferPos > 32) {
			memset(buffer + bufferPos, 0, 64 - bufferPos);
			processBuffer();
			bufferPos = 0;
		}
		memset(buffer + bufferPos, 0, 60 - bufferPos);

		((unsigned long*)buffer)[15] = _bswap(value);
		processBuffer();
		__stosb(buffer, 0, sizeof buffer);
	}
}
